/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */
package org.phenotips.vocabulary.internal.solr;

import org.phenotips.vocabulary.Vocabulary;
import org.phenotips.vocabulary.VocabularyInputTerm;
import org.phenotips.vocabulary.VocabularyTerm;

import java.util.Set;

import org.apache.solr.common.SolrInputDocument;
import org.apache.solr.common.SolrInputField;

/**
 * Implementation for {@link VocabularyInputTerm} based on a Solr document.
 *
 * @version $Id: 0e0d34471c42e9afc56dbfe5a548930435aec5c1 $
 */
public class SolrVocabularyInputTerm extends AbstractSolrVocabularyTerm implements VocabularyInputTerm
{
    /**
     * Constructor.
     *
     * @param doc the solr document representing the term
     * @param vocabulary the owner vocabulary
     */
    public SolrVocabularyInputTerm(SolrInputDocument doc, Vocabulary vocabulary)
    {
        super(doc, vocabulary);
        initialize();
    }

    @Override
    public VocabularyInputTerm setId(String id)
    {
        set(ID_KEY, id);
        return this;
    }

    @Override
    public VocabularyInputTerm setName(String name)
    {
        set(NAME_KEY, name);
        return this;
    }

    @Override
    public VocabularyInputTerm setDescription(String description)
    {
        set(DESCRIPTION_KEY, description);
        return this;
    }

    @Override
    public VocabularyInputTerm setParents(Set<VocabularyTerm> parents)
    {
        if (!isNull()) {
            SolrInputField field = new SolrInputField(PARENTS_KEY);
            for (VocabularyTerm parent : parents) {
                field.addValue(parent.getId(), 1.0f);
            }
            ((SolrInputDocument) this.doc).put(PARENTS_KEY, field);
        }
        return this;
    }

    @Override
    public VocabularyInputTerm set(String key, Object value)
    {
        if (!isNull()) {
            this.doc.setField(key, value);
        }
        return this;
    }

    @Override
    public VocabularyInputTerm append(String key, Object value)
    {
        if (!isNull()) {
            this.doc.addField(key, value);
        }
        return this;
    }

    @Override
    public Object get(String key)
    {
        // We have to override this because in Solr 5.5 SolrInputDocument#get wrongly forwards to #getFirstValue
        if (isNull()) {
            return null;
        }
        SolrInputField field = ((SolrInputDocument) this.doc).getField(key);
        if (field == null) {
            return null;
        }
        return field.getValue();
    }

    @Override
    public Set<VocabularyTerm> getParents()
    {
        // We have to override this because the parents might be re-set, so the result can't be cached
        return new LazySolrTermSet(getValues(PARENTS_KEY), this.vocabulary);
    }

    @Override
    public Set<VocabularyTerm> getAncestors()
    {
        // We have to override this because the ancestors might be re-set, so the result can't be cached
        return new LazySolrTermSet(getValues(ANCESTORS_KEY), this.vocabulary);
    }

    @Override
    public Set<VocabularyTerm> getAncestorsAndSelf()
    {
        // We have to override this because the ancestors might be re-set, so the result can't be cached
        return getUncachedAncestorsAndSelf();
    }
}
